@@ -14,17 +14,20 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
+enum got_diff_algorithm {
+	GOT_DIFF_ALGORITHM_MYERS,
+	GOT_DIFF_ALGORITHM_PATIENCE,
+};
+
 /*
  * Compute the differences between two blobs and write unified diff text
  * to the provided output file. Two open temporary files must be provided
  * for internal use; these files can be obtained from got_opentemp() and
  * must be closed by the caller.
  * If one of the blobs being diffed does not exist, all corresponding
- * blob object and temporary file arguments should be set to NULL.
+ * blob object arguments should be set to NULL.
  * Two const char * diff header labels may be provided which will be used
  * to identify each blob in the diff output.
- * The set of arguments relating to either blob may be NULL to indicate
- * that no content is present on its respective side of the diff.
  * If a label is NULL, use the blob's SHA1 checksum instead.
  * The number of context lines to show in the diff must be specified as well.
  * Whitespace differences may optionally be ignored.
@@ -33,19 +36,21 @@
  */
 const struct got_error *got_diff_blob(off_t **, size_t *,
     struct got_blob_object *, struct got_blob_object *, FILE *, FILE *,
-    const char *, const char *, int, int, int, FILE *);
+    const char *, const char *, enum got_diff_algorithm, int, int, int,
+    FILE *);
 
 /*
  * Compute the differences between a blob and a file and write unified diff
  * text to the provided output file. The blob object, its content, and its
- * size must be provided.The file's size must be provided, as well as a
+ * size must be provided. The file's size must be provided, as well as a
  * const char * diff header label which identifies the file.
  * An optional const char * diff header label for the blob may be provided, too.
  * The number of context lines to show in the diff must be specified as well.
  * Whitespace differences may optionally be ignored.
  */
 const struct got_error *got_diff_blob_file(struct got_blob_object *, FILE *,
-    off_t, const char *, FILE *, size_t, const char *, int, int, int, FILE *);
+    off_t, const char *, FILE *, int, size_t, const char *,
+    enum got_diff_algorithm, int, int, int, FILE *);
 
 /*
  * A callback function invoked to handle the differences between two blobs
@@ -55,7 +60,7 @@ const struct got_error *got_diff_blob_file(struct got_blob_object *, FILE *,
  * the second blob contains content on the new side of the diff.
  * Two open temporary files must be provided for internal use; these files
  * can be obtained from got_opentemp() and must be closed by the caller.
- * The set of arguments relating to either blob may be NULL to indicate
+ * The blob object argument for either blob may be NULL to indicate
  * that no content is present on its respective side of the diff.
  * File modes from relevant tree objects which contain the blobs may
  * also be passed. These will be zero if not available.
@@ -75,6 +80,7 @@ struct got_diff_blob_output_unidiff_arg {
 	int diff_context;	/* Sets the number of context lines. */
 	int ignore_whitespace;	/* Ignore whitespace differences. */
 	int force_text_diff;	/* Assume text even if binary data detected. */
+	enum got_diff_algorithm diff_algo; /* Diffing algorithm to use. */
 
 	/*
 	 * The number of lines contained in produced unidiff text output,
@@ -102,14 +108,16 @@ const struct got_error *got_diff_blob_output_unidiff(void *,
  * Diffing of blob content can be suppressed by passing zero for the
  * 'diff_content' parameter. The callback will then only receive blob
  * object IDs and diff labels, but NULL pointers instead of blob objects.
- * If 'diff_content' is set, two open temporary files must be provided
- * for internal use; these files can be obtained from got_opentemp()
+ * If 'diff_content' is set, two open temporary FILEs and two open
+ * temporary file descriptors must be provided for internal use; these
+ * files can be obtained from got_opentemp() and got_opentempfd(),
  * and must be closed by the caller. Otherwise the files can be NULL.
  * The set of arguments relating to either tree may be NULL to indicate
  * that no content is present on its respective side of the diff.
  */
 const struct got_error *got_diff_tree(struct got_tree_object *,
-    struct got_tree_object *, FILE *, FILE *, const char *, const char *,
+    struct got_tree_object *, FILE *, FILE *, int, int,
+    const char *, const char *,
     struct got_repository *, got_diff_blob_cb cb, void *cb_arg, int);
 
 /*
@@ -137,9 +145,10 @@ const struct got_error *got_diff_tree_collect_changed_paths(void *,
  * Diff two objects, assuming both objects are blobs. Two const char * diff
  * header labels may be provided which will be used to identify each blob in
  * the diff output. If a label is NULL, use the blob's SHA1 checksum instead.
- * Two open temporary files must be provided for internal use; these files
- * can be obtained from got_opentemp() and must be closed by the caller.
- * The set of arguments relating to either blob may be NULL to indicate
+ * Two open temporary files and two temporary file descriptors must be
+ * provided for internal use; these files can be obtained from
+ * got_opentemp() and got_opentempfd(), and must be closed by the caller.
+ * The set of arguments relating to either blob may be NULL/-1 to indicate
  * that no content is present on its respective side of the diff.
  * The number of context lines to show in the diff must be specified as well.
  * Write unified diff text to the provided output FILE.
@@ -147,17 +156,21 @@ const struct got_error *got_diff_tree_collect_changed_paths(void *,
  * array of line offsets for, and the number of lines in, the unidiff text.
  */
 const struct got_error *got_diff_objects_as_blobs(off_t **, size_t *,
-    FILE *, FILE *, struct got_object_id *, struct got_object_id *,
-    const char *, const char *, int, int, int,
+    FILE *, FILE *, int, int, struct got_object_id *, struct got_object_id *,
+    const char *, const char *, enum got_diff_algorithm, int, int, int,
     struct got_repository *, FILE *);
 
+struct got_pathlist_head;
+
 /*
  * Diff two objects, assuming both objects are trees. Two const char * diff
  * header labels may be provided which will be used to identify each blob in
  * the trees. If a label is NULL, use the blob's SHA1 checksum instead.
  * The number of context lines to show in diffs must be specified.
- * Two open temporary files must be provided for internal use; these files
- * can be obtained from got_opentemp() and must be closed by the caller.
+ * Two open temporary files and two temporary file descriptors must be
+ * provided for internal use; these files can be obtained from
+ * got_opentemp() and got_opentempfd(), and must be closed by the caller.
+ * If 'diff_content' is not set, the files may be NULL / -1.
  * The set of arguments relating to either tree may be NULL to indicate
  * that no content is present on its respective side of the diff.
  * Write unified diff text to the provided output FILE.
@@ -165,15 +178,16 @@ const struct got_error *got_diff_objects_as_blobs(off_t **, size_t *,
  * array of line offsets for, and the number of lines in, the unidiff text.
  */
 const struct got_error *got_diff_objects_as_trees(off_t **, size_t *,
-    FILE *, FILE *, struct got_object_id *, struct got_object_id *,
-    struct got_pathlist_head *, char *, char *, int, int, int,
-    struct got_repository *, FILE *);
+    FILE *, FILE *, int, int, struct got_object_id *, struct got_object_id *,
+    struct got_pathlist_head *, const char *, const char *,
+    enum got_diff_algorithm, int, int, int, struct got_repository *, FILE *);
 
 /*
  * Diff two objects, assuming both objects are commits.
  * The number of context lines to show in diffs must be specified.
- * Two open temporary files must be provided for internal use; these files
- * can be obtained from got_opentemp() and must be closed by the caller.
+ * Two open temporary files and two temporary file descriptors must be
+ * provided for internal use; these files can be obtained from
+ * got_opentemp() and got_opentempfd(), and must be closed by the caller.
  * The set of arguments relating to either commit may be NULL to indicate
  * that no content is present on its respective side of the diff.
  * Write unified diff text to the provided output FILE.
@@ -181,7 +195,8 @@ const struct got_error *got_diff_objects_as_trees(off_t **, size_t *,
  * array of line offsets for, and the number of lines in, the unidiff text.
  */
 const struct got_error *got_diff_objects_as_commits(off_t **, size_t *,
-    FILE *, FILE *, struct got_object_id *, struct got_object_id *,
-    struct got_pathlist_head *, int, int, int, struct got_repository *, FILE *);
+    FILE *, FILE *, int, int, struct got_object_id *, struct got_object_id *,
+    struct got_pathlist_head *, enum got_diff_algorithm, int, int, int,
+    struct got_repository *, FILE *);
 
 #define GOT_DIFF_MAX_CONTEXT	64
