@@ -17,12 +17,12 @@
 
 #include <stdlib.h>
 #include <string.h>
+#include <stdint.h>
 #include <stdio.h>
 #include <zlib.h>
 #include <limits.h>
 #include <time.h>
 #include <errno.h>
-#include <siphash.h>
 
 #include "got_compat.h"
 
