@@ -57,12 +57,13 @@ diff_blobs(off_t **line_offsets, size_t *nlines,
     struct got_diffreg_result **resultp, struct got_blob_object *blob1,
     struct got_blob_object *blob2, FILE *f1, FILE *f2,
     const char *label1, const char *label2, mode_t mode1, mode_t mode2,
-    int diff_context, int ignore_whitespace, int force_text_diff, FILE *outfile)
+    int diff_context, int ignore_whitespace, int force_text_diff, FILE *outfile,
+    enum got_diff_algorithm diff_algo)
 {
 	const struct got_error *err = NULL, *free_err;
 	char hex1[SHA1_DIGEST_STRING_LENGTH];
 	char hex2[SHA1_DIGEST_STRING_LENGTH];
-	char *idstr1 = NULL, *idstr2 = NULL;
+	const char *idstr1 = NULL, *idstr2 = NULL;
 	off_t size1, size2;
 	struct got_diffreg_result *result;
 	off_t outoff = 0;
@@ -160,8 +161,8 @@ diff_blobs(off_t **line_offsets, size_t *nlines,
 		free(modestr1);
 		free(modestr2);
 	}
-	err = got_diffreg(&result, f1, f2, GOT_DIFF_ALGORITHM_PATIENCE,
-	    ignore_whitespace, force_text_diff);
+	err = got_diffreg(&result, f1, f2, diff_algo, ignore_whitespace,
+	     force_text_diff);
 	if (err)
 		goto done;
 
@@ -197,30 +198,31 @@ got_diff_blob_output_unidiff(void *arg, struct got_blob_object *blob1,
 
 	return diff_blobs(&a->line_offsets, &a->nlines, NULL,
 	    blob1, blob2, f1, f2, label1, label2, mode1, mode2, a->diff_context,
-	    a->ignore_whitespace, a->force_text_diff, a->outfile);
+	    a->ignore_whitespace, a->force_text_diff, a->outfile, a->diff_algo);
 }
 
 const struct got_error *
 got_diff_blob(off_t **line_offsets, size_t *nlines,
     struct got_blob_object *blob1, struct got_blob_object *blob2,
     FILE *f1, FILE *f2, const char *label1, const char *label2,
-    int diff_context, int ignore_whitespace, int force_text_diff,
-    FILE *outfile)
+    enum got_diff_algorithm diff_algo, int diff_context,
+    int ignore_whitespace, int force_text_diff, FILE *outfile)
 {
 	return diff_blobs(line_offsets, nlines, NULL, blob1, blob2, f1, f2,
 	    label1, label2, 0, 0, diff_context, ignore_whitespace,
-	    force_text_diff, outfile);
+	    force_text_diff, outfile, diff_algo);
 }
 
 static const struct got_error *
 diff_blob_file(struct got_diffreg_result **resultp,
     struct got_blob_object *blob1, FILE *f1, off_t size1, const char *label1,
-    FILE *f2, size_t size2, const char *label2, int diff_context,
-    int ignore_whitespace, int force_text_diff, FILE *outfile)
+    FILE *f2, int f2_exists, size_t size2, const char *label2,
+    enum got_diff_algorithm diff_algo, int diff_context, int ignore_whitespace,
+    int force_text_diff, FILE *outfile)
 {
 	const struct got_error *err = NULL, *free_err;
 	char hex1[SHA1_DIGEST_STRING_LENGTH];
-	char *idstr1 = NULL;
+	const char *idstr1 = NULL;
 	struct got_diffreg_result *result = NULL;
 
 	if (resultp)
@@ -234,17 +236,17 @@ diff_blob_file(struct got_diffreg_result **resultp,
 	if (outfile) {
 		fprintf(outfile, "blob - %s\n", label1 ? label1 : idstr1);
 		fprintf(outfile, "file + %s\n",
-		    f2 == NULL ? "/dev/null" : label2);
+		    f2_exists ? label2 : "/dev/null");
 	}
 
-	err = got_diffreg(&result, f1, f2, GOT_DIFF_ALGORITHM_PATIENCE, 
-	    ignore_whitespace, force_text_diff);
+	err = got_diffreg(&result, f1, f2, diff_algo, ignore_whitespace,
+	    force_text_diff);
 	if (err)
 		goto done;
 
 	if (outfile) {
 		err = got_diffreg_output(NULL, NULL, result,
-		    f1 != NULL, f2 != NULL,
+		    blob1 != NULL, f2_exists,
 		    label2, /* show local file's path, not a blob ID */
 		    label2, GOT_DIFF_OUTPUT_UNIDIFF,
 		    diff_context, outfile);
@@ -265,16 +267,18 @@ done:
 
 const struct got_error *
 got_diff_blob_file(struct got_blob_object *blob1, FILE *f1, off_t size1,
-    const char *label1, FILE *f2, size_t size2, const char *label2,
-    int diff_context, int ignore_whitespace, int force_text_diff, FILE *outfile)
+    const char *label1, FILE *f2, int f2_exists, size_t size2,
+    const char *label2, enum got_diff_algorithm diff_algo, int diff_context,
+    int ignore_whitespace, int force_text_diff, FILE *outfile)
 {
-	return diff_blob_file(NULL, blob1, f1, size1, label1, f2, size2, label2,
-	    diff_context, ignore_whitespace, force_text_diff, outfile);
+	return diff_blob_file(NULL, blob1, f1, size1, label1, f2, f2_exists,
+	    size2, label2, diff_algo, diff_context, ignore_whitespace,
+	    force_text_diff, outfile );
 }
 
 static const struct got_error *
-diff_added_blob(struct got_object_id *id, FILE *f, const char *label,
-    mode_t mode, struct got_repository *repo,
+diff_added_blob(struct got_object_id *id, FILE *f1, FILE *f2, int fd2,
+    const char *label, mode_t mode, struct got_repository *repo,
     got_diff_blob_cb cb, void *cb_arg)
 {
 	const struct got_error *err;
@@ -285,10 +289,10 @@ diff_added_blob(struct got_object_id *id, FILE *f, const char *label,
 	if (err)
 		return err;
 
-	err = got_object_blob_open(&blob, repo, obj, 8192);
+	err = got_object_blob_open(&blob, repo, obj, 8192, fd2);
 	if (err)
 		goto done;
-	err = cb(cb_arg, NULL, blob, NULL, f, NULL, id,
+	err = cb(cb_arg, NULL, blob, f1, f2, NULL, id,
 	    NULL, label, 0, mode, repo);
 done:
 	got_object_close(obj);
@@ -299,7 +303,8 @@ done:
 
 static const struct got_error *
 diff_modified_blob(struct got_object_id *id1, struct got_object_id *id2,
-    FILE *f1, FILE *f2, const char *label1, const char *label2,
+    FILE *f1, FILE *f2, int fd1, int fd2,
+    const char *label1, const char *label2,
     mode_t mode1, mode_t mode2, struct got_repository *repo,
     got_diff_blob_cb cb, void *cb_arg)
 {
@@ -312,6 +317,7 @@ diff_modified_blob(struct got_object_id *id1, struct got_object_id *id2,
 	err = got_object_open(&obj1, repo, id1);
 	if (err)
 		return err;
+
 	if (obj1->type != GOT_OBJ_TYPE_BLOB) {
 		err = got_error(GOT_ERR_OBJ_TYPE);
 		goto done;
@@ -325,11 +331,11 @@ diff_modified_blob(struct got_object_id *id1, struct got_object_id *id2,
 		goto done;
 	}
 
-	err = got_object_blob_open(&blob1, repo, obj1, 8192);
+	err = got_object_blob_open(&blob1, repo, obj1, 8192, fd1);
 	if (err)
 		goto done;
 
-	err = got_object_blob_open(&blob2, repo, obj2, 8192);
+	err = got_object_blob_open(&blob2, repo, obj2, 8192, fd2);
 	if (err)
 		goto done;
 
@@ -348,8 +354,9 @@ done:
 }
 
 static const struct got_error *
-diff_deleted_blob(struct got_object_id *id, FILE *f, const char *label,
-    mode_t mode, struct got_repository *repo, got_diff_blob_cb cb, void *cb_arg)
+diff_deleted_blob(struct got_object_id *id, FILE *f1, int fd1,
+   FILE *f2, const char *label, mode_t mode, struct got_repository *repo,
+   got_diff_blob_cb cb, void *cb_arg)
 {
 	const struct got_error *err;
 	struct got_blob_object  *blob = NULL;
@@ -359,10 +366,10 @@ diff_deleted_blob(struct got_object_id *id, FILE *f, const char *label,
 	if (err)
 		return err;
 
-	err = got_object_blob_open(&blob, repo, obj, 8192);
+	err = got_object_blob_open(&blob, repo, obj, 8192, fd1);
 	if (err)
 		goto done;
-	err = cb(cb_arg, blob, NULL, f, NULL, id, NULL, label, NULL,
+	err = cb(cb_arg, blob, NULL, f1, f2, id, NULL, label, NULL,
 	    mode, 0, repo);
 done:
 	got_object_close(obj);
@@ -372,9 +379,9 @@ done:
 }
 
 static const struct got_error *
-diff_added_tree(struct got_object_id *id, FILE *f, const char *label,
-    struct got_repository *repo, got_diff_blob_cb cb, void *cb_arg,
-    int diff_content)
+diff_added_tree(struct got_object_id *id, FILE *f1, FILE *f2, int fd2,
+    const char *label, struct got_repository *repo, got_diff_blob_cb cb,
+    void *cb_arg, int diff_content)
 {
 	const struct got_error *err = NULL;
 	struct got_object *treeobj = NULL;
@@ -393,7 +400,7 @@ diff_added_tree(struct got_object_id *id, FILE *f, const char *label,
 	if (err)
 		goto done;
 
-	err = got_diff_tree(NULL, tree, NULL, f, NULL, label,
+	err = got_diff_tree(NULL, tree, f1, f2, -1, fd2, NULL, label,
 	    repo, cb, cb_arg, diff_content);
 done:
 	if (tree)
@@ -405,7 +412,8 @@ done:
 
 static const struct got_error *
 diff_modified_tree(struct got_object_id *id1, struct got_object_id *id2,
-    FILE *f1, FILE *f2, const char *label1, const char *label2,
+    FILE *f1, FILE *f2, int fd1, int fd2,
+    const char *label1, const char *label2,
     struct got_repository *repo, got_diff_blob_cb cb, void *cb_arg,
     int diff_content)
 {
@@ -441,8 +449,8 @@ diff_modified_tree(struct got_object_id *id1, struct got_object_id *id2,
 	if (err)
 		goto done;
 
-	err = got_diff_tree(tree1, tree2, f1, f2, label1, label2,
-	    repo, cb, cb_arg, diff_content);
+	err = got_diff_tree(tree1, tree2, f1, f2, fd1, fd2,
+	    label1, label2, repo, cb, cb_arg, diff_content);
 
 done:
 	if (tree1)
@@ -457,9 +465,9 @@ done:
 }
 
 static const struct got_error *
-diff_deleted_tree(struct got_object_id *id, FILE *f, const char *label,
-    struct got_repository *repo, got_diff_blob_cb cb, void *cb_arg,
-    int diff_content)
+diff_deleted_tree(struct got_object_id *id, FILE *f1, int fd1,
+    FILE *f2, const char *label, struct got_repository *repo,
+    got_diff_blob_cb cb, void *cb_arg, int diff_content)
 {
 	const struct got_error *err;
 	struct got_object *treeobj = NULL;
@@ -478,7 +486,7 @@ diff_deleted_tree(struct got_object_id *id, FILE *f, const char *label,
 	if (err)
 		goto done;
 
-	err = got_diff_tree(tree, NULL, f, NULL, label, NULL,
+	err = got_diff_tree(tree, NULL, f1, f2, fd1, -1, label, NULL,
 	    repo, cb, cb_arg, diff_content);
 done:
 	if (tree)
@@ -499,7 +507,8 @@ diff_kind_mismatch(struct got_object_id *id1, struct got_object_id *id2,
 
 static const struct got_error *
 diff_entry_old_new(struct got_tree_entry *te1, struct got_tree_entry *te2,
-    FILE *f1, FILE *f2, const char *label1, const char *label2,
+    FILE *f1, FILE *f2, int fd1, int fd2,
+    const char *label1, const char *label2,
     struct got_repository *repo, got_diff_blob_cb cb, void *cb_arg,
     int diff_content)
 {
@@ -511,12 +520,12 @@ diff_entry_old_new(struct got_tree_entry *te1, struct got_tree_entry *te2,
 
 	if (te2 == NULL) {
 		if (S_ISDIR(te1->mode))
-			err = diff_deleted_tree(&te1->id, f1, label1, repo,
-			    cb, cb_arg, diff_content);
+			err = diff_deleted_tree(&te1->id, f1, fd1, f2,
+			    label1, repo, cb, cb_arg, diff_content);
 		else {
 			if (diff_content)
-				err = diff_deleted_blob(&te1->id, f1, label1,
-				    te1->mode, repo, cb, cb_arg);
+				err = diff_deleted_blob(&te1->id, f1, fd1,
+				    f2, label1, te1->mode, repo, cb, cb_arg);
 			else
 				err = cb(cb_arg, NULL, NULL, NULL, NULL,
 				    &te1->id, NULL, label1, NULL,
@@ -530,7 +539,8 @@ diff_entry_old_new(struct got_tree_entry *te1, struct got_tree_entry *te2,
 	if (S_ISDIR(te1->mode) && S_ISDIR(te2->mode)) {
 		if (!id_match)
 			return diff_modified_tree(&te1->id, &te2->id, f1, f2,
-			    label1, label2, repo, cb, cb_arg, diff_content);
+			    fd1, fd2, label1, label2, repo, cb, cb_arg,
+			    diff_content);
 	} else if ((S_ISREG(te1->mode) || S_ISLNK(te1->mode)) &&
 	    (S_ISREG(te2->mode) || S_ISLNK(te2->mode))) {
 		if (!id_match ||
@@ -538,8 +548,8 @@ diff_entry_old_new(struct got_tree_entry *te1, struct got_tree_entry *te2,
 		    (te2->mode & (S_IFLNK | S_IXUSR))) {
 			if (diff_content)
 				return diff_modified_blob(&te1->id, &te2->id,
-				    f1, f2, label1, label2, te1->mode,
-				    te2->mode, repo, cb, cb_arg);
+				    f1, f2, fd1, fd2, label1, label2,
+				    te1->mode, te2->mode, repo, cb, cb_arg);
 			else
 				return cb(cb_arg, NULL, NULL, NULL, NULL,
 				    &te1->id, &te2->id, label1, label2,
@@ -556,7 +566,7 @@ diff_entry_old_new(struct got_tree_entry *te1, struct got_tree_entry *te2,
 
 static const struct got_error *
 diff_entry_new_old(struct got_tree_entry *te2,
-    struct got_tree_entry *te1, FILE *f2, const char *label2,
+    struct got_tree_entry *te1, FILE *f1, FILE *f2, int fd2, const char *label2,
     struct got_repository *repo, got_diff_blob_cb cb, void *cb_arg,
     int diff_content)
 {
@@ -567,12 +577,12 @@ diff_entry_new_old(struct got_tree_entry *te2,
 		return NULL;
 
 	if (S_ISDIR(te2->mode))
-		return diff_added_tree(&te2->id, f2, label2,
+		return diff_added_tree(&te2->id, f1, f2, fd2, label2,
 		    repo, cb, cb_arg, diff_content);
 
 	if (diff_content)
-		return diff_added_blob(&te2->id, f2, label2, te2->mode,
-		    repo, cb, cb_arg);
+		return diff_added_blob(&te2->id, f1, f2, fd2,
+		    label2, te2->mode, repo, cb, cb_arg);
 
 	return cb(cb_arg, NULL, NULL, NULL, NULL, NULL, &te2->id,
 	    NULL, label2, 0, te2->mode, repo);
@@ -623,7 +633,8 @@ done:
 
 const struct got_error *
 got_diff_tree(struct got_tree_object *tree1, struct got_tree_object *tree2,
-    FILE *f1, FILE *f2, const char *label1, const char *label2,
+    FILE *f1, FILE *f2, int fd1, int fd2,
+    const char *label1, const char *label2,
     struct got_repository *repo, got_diff_blob_cb cb, void *cb_arg,
     int diff_content)
 {
@@ -660,8 +671,8 @@ got_diff_tree(struct got_tree_object *tree1, struct got_tree_object *tree2,
 					return
 					    got_error_from_errno("asprintf");
 			}
-			err = diff_entry_old_new(te1, te, f1, f2, l1, l2,
-			    repo, cb, cb_arg, diff_content);
+			err = diff_entry_old_new(te1, te, f1, f2, fd1, fd2,
+			    l1, l2, repo, cb, cb_arg, diff_content);
 			if (err)
 				break;
 		}
@@ -683,8 +694,8 @@ got_diff_tree(struct got_tree_object *tree1, struct got_tree_object *tree2,
 					return
 					    got_error_from_errno("asprintf");
 			}
-			err = diff_entry_new_old(te2, te, f2, l2, repo,
-			    cb, cb_arg, diff_content);
+			err = diff_entry_new_old(te2, te, f1, f2, fd2, l2,
+			    repo, cb, cb_arg, diff_content);
 			if (err)
 				break;
 		}
@@ -716,8 +727,10 @@ got_diff_tree(struct got_tree_object *tree1, struct got_tree_object *tree2,
 
 const struct got_error *
 got_diff_objects_as_blobs(off_t **line_offsets, size_t *nlines,
-    FILE *f1, FILE *f2, struct got_object_id *id1, struct got_object_id *id2,
-    const char *label1, const char *label2, int diff_context,
+    FILE *f1, FILE *f2, int fd1, int fd2,
+    struct got_object_id *id1, struct got_object_id *id2,
+    const char *label1, const char *label2,
+    enum got_diff_algorithm diff_algo, int diff_context,
     int ignore_whitespace, int force_text_diff,
     struct got_repository *repo, FILE *outfile)
 {
@@ -728,18 +741,18 @@ got_diff_objects_as_blobs(off_t **line_offsets, size_t *nlines,
 		return got_error(GOT_ERR_NO_OBJ);
 
 	if (id1) {
-		err = got_object_open_as_blob(&blob1, repo, id1, 8192);
+		err = got_object_open_as_blob(&blob1, repo, id1, 8192, fd1);
 		if (err)
 			goto done;
 	}
 	if (id2) {
-		err = got_object_open_as_blob(&blob2, repo, id2, 8192);
+		err = got_object_open_as_blob(&blob2, repo, id2, 8192, fd2);
 		if (err)
 			goto done;
 	}
 	err = got_diff_blob(line_offsets, nlines, blob1, blob2, f1, f2,
-	    label1, label2, diff_context, ignore_whitespace, force_text_diff,
-	    outfile);
+	    label1, label2, diff_algo, diff_context, ignore_whitespace,
+	    force_text_diff, outfile);
 done:
 	if (blob1)
 		got_object_blob_close(blob1);
@@ -750,7 +763,7 @@ done:
 
 static const struct got_error *
 diff_paths(struct got_tree_object *tree1, struct got_tree_object *tree2,
-    FILE *f1, FILE *f2, struct got_pathlist_head *paths,
+    FILE *f1, FILE *f2, int fd1, int fd2, struct got_pathlist_head *paths,
     struct got_repository *repo, got_diff_blob_cb cb, void *cb_arg)
 {
 	const struct got_error *err = NULL;
@@ -820,13 +833,13 @@ diff_paths(struct got_tree_object *tree1, struct got_tree_object *tree2,
 		    type2 == GOT_OBJ_TYPE_BLOB) {
 			if (id1) {
 				err = got_object_open_as_blob(&blob1, repo,
-				    id1, 8192);
+				    id1, 8192, fd1);
 				if (err)
 					goto done;
 			}
 			if (id2) {
 				err = got_object_open_as_blob(&blob2, repo,
-				    id2, 8192);
+				    id2, 8192, fd2);
 				if (err)
 					goto done;
 			}
@@ -851,6 +864,7 @@ diff_paths(struct got_tree_object *tree1, struct got_tree_object *tree2,
 					goto done;
 			}
 			err = got_diff_tree(subtree1, subtree2, f1, f2,
+			    fd1, fd2,
 			    id1 ? pe->path : "/dev/null",
 			    id2 ? pe->path : "/dev/null",
 			    repo, cb, cb_arg, 1);
@@ -875,12 +889,40 @@ done:
 	return err;
 }
 
-const struct got_error *
-got_diff_objects_as_trees(off_t **line_offsets, size_t *nlines,
-    FILE *f1, FILE *f2, struct got_object_id *id1, struct got_object_id *id2,
-    struct got_pathlist_head *paths,
-    char *label1, char *label2, int diff_context, int ignore_whitespace,
-    int force_text_diff, struct got_repository *repo, FILE *outfile)
+static const struct got_error *
+show_object_id(off_t **line_offsets, size_t *nlines, const char *obj_typestr,
+    int ch, const char *id_str, FILE *outfile)
+{
+	const struct got_error *err;
+	int n;
+	off_t outoff = 0;
+
+	n = fprintf(outfile, "%s %c %s\n", obj_typestr, ch, id_str);
+	if (line_offsets != NULL && *line_offsets != NULL) {
+		if (*nlines == 0) {
+			err = add_line_offset(line_offsets, nlines, 0);
+			if (err)
+				return err;
+		} else
+			outoff = (*line_offsets)[*nlines - 1];
+
+		outoff += n;
+		err = add_line_offset(line_offsets, nlines, outoff);
+		if (err)
+			return err;
+	}
+
+	return NULL;
+}
+
+static const struct got_error *
+diff_objects_as_trees(off_t **line_offsets, size_t *nlines,
+    FILE *f1, FILE *f2, int fd1, int fd2,
+    struct got_object_id *id1, struct got_object_id *id2,
+    struct got_pathlist_head *paths, const char *label1, const char *label2,
+    int diff_context, int ignore_whitespace, int force_text_diff,
+    struct got_repository *repo, FILE *outfile,
+    enum got_diff_algorithm diff_algo)
 {
 	const struct got_error *err;
 	struct got_tree_object *tree1 = NULL, *tree2 = NULL;
@@ -901,6 +943,7 @@ got_diff_objects_as_trees(off_t **line_offsets, size_t *nlines,
 			goto done;
 	}
 
+	arg.diff_algo = diff_algo;
 	arg.diff_context = diff_context;
 	arg.ignore_whitespace = ignore_whitespace;
 	arg.force_text_diff = force_text_diff;
@@ -913,10 +956,11 @@ got_diff_objects_as_trees(off_t **line_offsets, size_t *nlines,
 		arg.nlines = 0;
 	}
 	if (paths == NULL || TAILQ_EMPTY(paths)) {
-		err = got_diff_tree(tree1, tree2, f1, f2, label1, label2,
-		    repo, got_diff_blob_output_unidiff, &arg, 1);
+		err = got_diff_tree(tree1, tree2, f1, f2, fd1, fd2,
+		    label1, label2, repo,
+		    got_diff_blob_output_unidiff, &arg, 1);
 	} else {
-		err = diff_paths(tree1, tree2, f1, f2, paths, repo,
+		err = diff_paths(tree1, tree2, f1, f2, fd1, fd2, paths, repo,
 		    got_diff_blob_output_unidiff, &arg);
 	}
 	if (want_lineoffsets) {
@@ -931,15 +975,73 @@ done:
 	return err;
 }
 
+const struct got_error *
+got_diff_objects_as_trees(off_t **line_offsets, size_t *nlines,
+    FILE *f1, FILE *f2, int fd1, int fd2,
+    struct got_object_id *id1, struct got_object_id *id2,
+    struct got_pathlist_head *paths, const char *label1, const char *label2,
+    enum got_diff_algorithm diff_algo, int diff_context, int ignore_whitespace,
+    int force_text_diff, struct got_repository *repo, FILE *outfile)
+{
+	const struct got_error *err;
+	char *idstr = NULL;
+
+	if (id1 == NULL && id2 == NULL)
+		return got_error(GOT_ERR_NO_OBJ);
+
+	if (id1) {
+		err = got_object_id_str(&idstr, id1);
+		if (err)
+			goto done;
+		err = show_object_id(line_offsets, nlines, "tree", '-',
+		    idstr, outfile);
+		if (err)
+			goto done;
+		free(idstr);
+		idstr = NULL;
+	} else {
+		err = show_object_id(line_offsets, nlines, "tree", '-',
+		    "/dev/null", outfile);
+		if (err)
+			goto done;
+	}
+
+	if (id2) {
+		err = got_object_id_str(&idstr, id2);
+		if (err)
+			goto done;
+		err = show_object_id(line_offsets, nlines, "tree", '+',
+		    idstr, outfile);
+		if (err)
+			goto done;
+		free(idstr);
+		idstr = NULL;
+	} else {
+		err = show_object_id(line_offsets, nlines, "tree", '+',
+		    "/dev/null", outfile);
+		if (err)
+			goto done;
+	}
+
+	err = diff_objects_as_trees(line_offsets, nlines, f1, f2, fd1, fd2,
+	    id1, id2, paths, label1, label2, diff_context, ignore_whitespace,
+	    force_text_diff, repo, outfile, diff_algo);
+done:
+	free(idstr);
+	return err;
+}
+
 const struct got_error *
 got_diff_objects_as_commits(off_t **line_offsets, size_t *nlines,
-    FILE *f1, FILE *f2, struct got_object_id *id1, struct got_object_id *id2,
-    struct got_pathlist_head *paths,
+    FILE *f1, FILE *f2, int fd1, int fd2,
+    struct got_object_id *id1, struct got_object_id *id2,
+    struct got_pathlist_head *paths, enum got_diff_algorithm diff_algo,
     int diff_context, int ignore_whitespace, int force_text_diff,
     struct got_repository *repo, FILE *outfile)
 {
 	const struct got_error *err;
 	struct got_commit_object *commit1 = NULL, *commit2 = NULL;
+	char *idstr = NULL;
 
 	if (id2 == NULL)
 		return got_error(GOT_ERR_NO_OBJ);
@@ -948,29 +1050,53 @@ got_diff_objects_as_commits(off_t **line_offsets, size_t *nlines,
 		err = got_object_open_as_commit(&commit1, repo, id1);
 		if (err)
 			goto done;
+		err = got_object_id_str(&idstr, id1);
+		if (err)
+			goto done;
+		err = show_object_id(line_offsets, nlines, "commit", '-',
+		    idstr, outfile);
+		if (err)
+			goto done;
+		free(idstr);
+		idstr = NULL;
+	} else {
+		err = show_object_id(line_offsets, nlines, "commit", '-',
+		    "/dev/null", outfile);
+		if (err)
+			goto done;
 	}
 
 	err = got_object_open_as_commit(&commit2, repo, id2);
 	if (err)
 		goto done;
 
-	err = got_diff_objects_as_trees(line_offsets, nlines, f1, f2,
+	err = got_object_id_str(&idstr, id2);
+	if (err)
+		goto done;
+	err = show_object_id(line_offsets, nlines, "commit", '+',
+	    idstr, outfile);
+	if (err)
+		goto done;
+
+	err = diff_objects_as_trees(line_offsets, nlines, f1, f2, fd1, fd2,
 	    commit1 ? got_object_commit_get_tree_id(commit1) : NULL,
 	    got_object_commit_get_tree_id(commit2), paths, "", "",
-	    diff_context, ignore_whitespace, force_text_diff, repo, outfile);
+	    diff_context, ignore_whitespace, force_text_diff, repo, outfile,
+	    diff_algo);
 done:
 	if (commit1)
 		got_object_commit_close(commit1);
 	if (commit2)
 		got_object_commit_close(commit2);
+	free(idstr);
 	return err;
 }
 
 const struct got_error *
 got_diff_files(struct got_diffreg_result **resultp,
-    FILE *f1, const char *label1, FILE *f2, const char *label2,
-    int diff_context, int ignore_whitespace, int force_text_diff,
-    FILE *outfile)
+    FILE *f1, int f1_exists, const char *label1, FILE *f2, int f2_exists,
+    const char *label2, int diff_context, int ignore_whitespace,
+    int force_text_diff, FILE *outfile, enum got_diff_algorithm diff_algo)
 {
 	const struct got_error *err = NULL;
 	struct got_diffreg_result *diffreg_result = NULL;
@@ -980,19 +1106,19 @@ got_diff_files(struct got_diffreg_result **resultp,
 
 	if (outfile) {
 		fprintf(outfile, "file - %s\n",
-		    f1 == NULL ? "/dev/null" : label1);
+		    f1_exists ? label1 : "/dev/null");
 		fprintf(outfile, "file + %s\n",
-		    f2 == NULL ? "/dev/null" : label2);
+		    f2_exists ? label2 : "/dev/null");
 	}
 
-	err = got_diffreg(&diffreg_result, f1, f2, GOT_DIFF_ALGORITHM_PATIENCE,
+	err = got_diffreg(&diffreg_result, f1, f2, diff_algo,
 	    ignore_whitespace, force_text_diff);
 	if (err)
 		goto done;
 
 	if (outfile) {
 		err = got_diffreg_output(NULL, NULL, diffreg_result,
-		    f1 != NULL, f2 != NULL, label1, label2,
+		    f1_exists, f2_exists, label1, label2,
 		    GOT_DIFF_OUTPUT_UNIDIFF, diff_context, outfile);
 		if (err)
 			goto done;
