@@ -174,6 +174,7 @@ enum got_imsg_type {
 	GOT_IMSG_GOTCONFIG_AUTHOR_REQUEST,
 	GOT_IMSG_GOTCONFIG_ALLOWEDSIGNERS_REQUEST,
 	GOT_IMSG_GOTCONFIG_REVOKEDSIGNERS_REQUEST,
+	GOT_IMSG_GOTCONFIG_SIGNERID_REQUEST,
 	GOT_IMSG_GOTCONFIG_REMOTES_REQUEST,
 	GOT_IMSG_GOTCONFIG_INT_VAL,
 	GOT_IMSG_GOTCONFIG_STR_VAL,
@@ -766,6 +767,8 @@ const struct got_error *got_privsep_send_gotconfig_allowed_signers_req(
     struct imsgbuf *);
 const struct got_error *got_privsep_send_gotconfig_revoked_signers_req(
     struct imsgbuf *);
+const struct got_error *got_privsep_send_gotconfig_signer_id_req(
+    struct imsgbuf *);
 const struct got_error *got_privsep_send_gotconfig_remotes_req(
     struct imsgbuf *);
 const struct got_error *got_privsep_recv_gotconfig_str(char **,
