@@ -709,9 +709,6 @@ got_privsep_recv_fetch_progress(int *done, struct got_object_id **id,
 
 	datalen = imsg.hdr.len - IMSG_HEADER_SIZE;
 	switch (imsg.hdr.type) {
-	case GOT_IMSG_ERROR:
-		err = recv_imsg_error(&imsg, datalen);
-		break;
 	case GOT_IMSG_FETCH_SYMREFS:
 		if (datalen < sizeof(*isymrefs)) {
 			err = got_error(GOT_ERR_PRIVSEP_LEN);
@@ -933,9 +930,6 @@ got_privsep_recv_send_remote_refs(struct got_pathlist_head *remote_refs,
 			return err;
 		datalen = imsg.hdr.len - IMSG_HEADER_SIZE;
 		switch (imsg.hdr.type) {
-		case GOT_IMSG_ERROR:
-			err = recv_imsg_error(&imsg, datalen);
-			goto done;
 		case GOT_IMSG_SEND_REMOTE_REF:
 			if (datalen < sizeof(iremote_ref)) {
 				err = got_error(GOT_ERR_PRIVSEP_MSG);
@@ -1016,9 +1010,6 @@ got_privsep_recv_send_progress(int *done, off_t *bytes_sent,
 
 	datalen = imsg.hdr.len - IMSG_HEADER_SIZE;
 	switch (imsg.hdr.type) {
-	case GOT_IMSG_ERROR:
-		err = recv_imsg_error(&imsg, datalen);
-		break;
 	case GOT_IMSG_SEND_UPLOAD_PROGRESS:
 		if (datalen < sizeof(*bytes_sent)) {
 			err = got_error(GOT_ERR_PRIVSEP_MSG);
@@ -1099,9 +1090,6 @@ got_privsep_recv_index_progress(int *done, int *nobj_total,
 
 	datalen = imsg.hdr.len - IMSG_HEADER_SIZE;
 	switch (imsg.hdr.type) {
-	case GOT_IMSG_ERROR:
-		err = recv_imsg_error(&imsg, datalen);
-		break;
 	case GOT_IMSG_IDXPACK_PROGRESS:
 		if (datalen < sizeof(*iprogress)) {
 			err = got_error(GOT_ERR_PRIVSEP_LEN);
@@ -1591,43 +1579,17 @@ got_privsep_recv_tree(struct got_tree_object **tree, struct imsgbuf *ibuf)
 
 	*tree = NULL;
 
-	err = read_imsg(ibuf);
-	if (err)
-		goto done;
-
-	for (;;) {
+	while (*tree == NULL || nentries < (*tree)->nentries) {
 		struct imsg imsg;
-		size_t n;
 		size_t datalen;
 
-		n = imsg_get(ibuf, &imsg);
-		if (n == 0) {
-			if ((*tree)) {
-				if (nentries < (*tree)->nentries) {
-					err = read_imsg(ibuf);
-					if (err)
-						break;
-					continue;
-				} else
-					break;
-			} else {
-				err = got_error(GOT_ERR_PRIVSEP_MSG);
-				break;
-			}
-		}
-
-		if (imsg.hdr.len < IMSG_HEADER_SIZE + min_datalen) {
-			imsg_free(&imsg);
-			err = got_error(GOT_ERR_PRIVSEP_LEN);
+		err = got_privsep_recv_imsg(&imsg, ibuf, min_datalen);
+		if (err)
 			break;
-		}
 
 		datalen = imsg.hdr.len - IMSG_HEADER_SIZE;
 
 		switch (imsg.hdr.type) {
-		case GOT_IMSG_ERROR:
-			err = recv_imsg_error(&imsg, datalen);
-			break;
 		case GOT_IMSG_TREE:
 			/* This message should only appear once. */
 			if (*tree != NULL) {
@@ -1677,7 +1639,7 @@ got_privsep_recv_tree(struct got_tree_object **tree, struct imsgbuf *ibuf)
 		if (err)
 			break;
 	}
-done:
+
 	if (*tree && (*tree)->nentries != nentries) {
 		if (err == NULL)
 			err = got_error(GOT_ERR_PRIVSEP_LEN);
@@ -2386,6 +2348,17 @@ got_privsep_send_gotconfig_revoked_signers_req(struct imsgbuf *ibuf)
 	return flush_imsg(ibuf);
 }
 
+const struct got_error *
+got_privsep_send_gotconfig_signer_id_req(struct imsgbuf *ibuf)
+{
+	if (imsg_compose(ibuf,
+	    GOT_IMSG_GOTCONFIG_SIGNERID_REQUEST, 0, 0, -1, NULL, 0) == -1)
+		return got_error_from_errno("imsg_compose "
+		    "GOTCONFIG_SIGNERID_REQUEST");
+
+	return flush_imsg(ibuf);
+}
+
 const struct got_error *
 got_privsep_send_gotconfig_remotes_req(struct imsgbuf *ibuf)
 {
@@ -2412,9 +2385,6 @@ got_privsep_recv_gotconfig_str(char **str, struct imsgbuf *ibuf)
 	datalen = imsg.hdr.len - IMSG_HEADER_SIZE;
 
 	switch (imsg.hdr.type) {
-	case GOT_IMSG_ERROR:
-		err = recv_imsg_error(&imsg, datalen);
-		break;
 	case GOT_IMSG_GOTCONFIG_STR_VAL:
 		if (datalen == 0)
 			break;
@@ -2458,9 +2428,6 @@ got_privsep_recv_gotconfig_remotes(struct got_remote_repo **remotes,
 	datalen = imsg.hdr.len - IMSG_HEADER_SIZE;
 
 	switch (imsg.hdr.type) {
-	case GOT_IMSG_ERROR:
-		err = recv_imsg_error(&imsg, datalen);
-		break;
 	case GOT_IMSG_GOTCONFIG_REMOTES:
 		if (datalen != sizeof(iremotes)) {
 			err = got_error(GOT_ERR_PRIVSEP_LEN);
@@ -2499,9 +2466,6 @@ got_privsep_recv_gotconfig_remotes(struct got_remote_repo **remotes,
 		datalen = imsg.hdr.len - IMSG_HEADER_SIZE;
 
 		switch (imsg.hdr.type) {
-		case GOT_IMSG_ERROR:
-			err = recv_imsg_error(&imsg, datalen);
-			break;
 		case GOT_IMSG_GOTCONFIG_REMOTE:
 			remote = &(*remotes)[*nremotes];
 			memset(remote, 0, sizeof(*remote));
